@@ -7,8 +7,6 @@ sub-modules, as described below.
 
 """
 
-# DO SOME OS-WISE PATCHES
-
 import datetime as _datetime
 import os as _os
 import platform as _platform
@@ -16,28 +14,10 @@ import signal as _signal
 import sys as _sys
 import types as _types
 
-from google.protobuf.internal import api_implementation as _api_implementation
-
-if _api_implementation._default_implementation_type != 'cpp':
-    import warnings as _warnings
-
-    _warnings.warn(
-        '''
-    You are using Python protobuf backend, not the C++ version, which is much faster.
-
-    This is often due to C++ implementation failed to compile while installing Protobuf
-    - You are using in Python 3.9 (https://github.com/jina-ai/jina/issues/1801)
-    - You are using on architecture other than x86_64/armv6/armv7
-    - You installation is broken, try `pip install --force protobuf`
-    - You have C++ backend but you shut it down, try `export PROTOCOL_BUFFERS_PYTHON_IMPLEMENTATION=cpp`
-
-    ''',
-        RuntimeWarning,
-    )
-
 if _sys.version_info < (3, 7, 0) or _sys.version_info >= (3, 10, 0):
     raise OSError(f'Jina requires Python 3.7/3.8/3.9, but yours is {_sys.version_info}')
 
+# DO SOME OS-WISE PATCHES
 if _sys.version_info >= (3, 8, 0) and _platform.system() == 'Darwin':
     # temporary fix for python 3.8 on macos where the default start is set to "spawn"
     # https://docs.python.org/3/library/multiprocessing.html#contexts-and-start-methods
@@ -48,19 +28,15 @@ if _sys.version_info >= (3, 8, 0) and _platform.system() == 'Darwin':
 # fix fork error on MacOS but seems no effect? must do EXPORT manually before jina start
 _os.environ['OBJC_DISABLE_INITIALIZE_FORK_SAFETY'] = 'YES'
 
-# Underscore variables shared globally
-
-__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
-__license__ = "Apache-2.0"
-
 # do not change this line manually
 # this is managed by git tag and updated on every release
 # NOTE: this represents the NEXT release version
-__version__ = '1.2.2'
+
+__version__ = '2.0.5'
 
 # do not change this line manually
 # this is managed by proto/build-proto.sh and updated on every execution
-__proto_version__ = '0.0.81'
+__proto_version__ = '0.0.84'
 
 __uptime__ = _datetime.datetime.now().isoformat()
 
@@ -70,75 +46,62 @@ __uptime__ = _datetime.datetime.now().isoformat()
 # 3. copy all lines EXCEPT the first (which is the grep command in the last line)
 __jina_env__ = (
     'JINA_ARRAY_QUANT',
-    'JINA_BINARY_DELIMITER',
-    'JINA_CONTRIB_MODULE',
-    'JINA_CONTRIB_MODULE_IS_LOADING',
     'JINA_CONTROL_PORT',
     'JINA_DEFAULT_HOST',
     'JINA_DISABLE_UVLOOP',
-    'JINA_EXECUTOR_WORKDIR',
     'JINA_FULL_CLI',
-    'JINA_IPC_SOCK_TMP',
+    'JINA_HUBBLE_REGISTRY',
+    'JINA_HUB_CACHE_DIR',
+    'JINA_HUB_ROOT',
     'JINA_LOG_CONFIG',
     'JINA_LOG_ID',
     'JINA_LOG_LEVEL',
     'JINA_LOG_NO_COLOR',
     'JINA_LOG_WORKSPACE',
+    'JINA_OPTIMIZER_TRIAL_WORKSPACE',
     'JINA_POD_NAME',
-    'JINA_RAISE_ERROR_EARLY',
-    'JINA_RANDOM_PORTS',
     'JINA_RANDOM_PORT_MAX',
     'JINA_RANDOM_PORT_MIN',
-    'JINA_SOCKET_HWM',
     'JINA_VCS_VERSION',
-    'JINA_WARN_UNNAMED',
-    'JINA_WORKSPACE',
 )
 
 __default_host__ = _os.environ.get('JINA_DEFAULT_HOST', '0.0.0.0')
+__docker_host__ = 'host.docker.internal'
+__default_executor__ = 'BaseExecutor'
+__default_endpoint__ = '/default'
 __ready_msg__ = 'ready and listening'
 __stop_msg__ = 'terminated'
-__binary_delimiter__ = _os.environ.get(
-    'JINA_BINARY_DELIMITER', '460841a0a8a430ae25d9ad7c1f048c57'
-).encode()
+__unset_msg__ = '(unset)'
+__num_args_executor_func__ = 5
 __root_dir__ = _os.path.dirname(_os.path.abspath(__file__))
+__resources_path__ = _os.path.join(
+    _os.path.dirname(_sys.modules['jina'].__file__), 'resources'
+)
+
 
 _names_with_underscore = [
     '__version__',
-    '__copyright__',
-    '__license__',
     '__proto_version__',
     '__default_host__',
     '__ready_msg__',
     '__stop_msg__',
-    '__binary_delimiter__',
     '__jina_env__',
     '__uptime__',
     '__root_dir__',
+    '__default_endpoint__',
+    '__default_executor__',
+    '__num_args_executor_func__',
+    '__unset_msg__',
 ]
 
-# Primitive data type,
-# note, they must be loaded BEFORE all executors/drivers/... to avoid cyclic imports
-from jina.types.ndarray.generic import NdArray
-from jina.types.request import Request, Response
-from jina.types.message import Message
-from jina.types.querylang import QueryLang
-from jina.types.document import Document
-from jina.types.document.multimodal import MultimodalDocument
-from jina.types.arrays import DocumentArray, QueryLangArray
 
 # ADD GLOBAL NAMESPACE VARIABLES
 JINA_GLOBAL = _types.SimpleNamespace()
 JINA_GLOBAL.scipy_installed = None
 JINA_GLOBAL.tensorflow_installed = None
 JINA_GLOBAL.torch_installed = None
+JINA_GLOBAL.dgl_installed = None
 
-import jina.importer as _ji
-
-# driver first, as executor may contain driver
-_ji.import_classes('jina.drivers', show_import_table=False, import_once=True)
-_ji.import_classes('jina.executors', show_import_table=False, import_once=True)
-_ji.import_classes('jina.hub', show_import_table=False, import_once=True)
 
 _signal.signal(_signal.SIGINT, _signal.default_int_handler)
 
@@ -158,8 +121,6 @@ def _set_nofile(nofile_atleast=4096):
     except ImportError:  # Windows
         res = None
 
-    from .logging import default_logger
-
     if res is None:
         return (None,) * 2
 
@@ -171,44 +132,38 @@ def _set_nofile(nofile_atleast=4096):
         if hard < soft:
             hard = soft
 
-        default_logger.debug(f'setting soft & hard ulimit -n {soft} {hard}')
         try:
             res.setrlimit(res.RLIMIT_NOFILE, (soft, hard))
         except (ValueError, res.error):
             try:
                 hard = soft
-                default_logger.warning(
-                    f'trouble with max limit, retrying with soft,hard {soft},{hard}'
-                )
+                print(f'trouble with max limit, retrying with soft,hard {soft},{hard}')
                 res.setrlimit(res.RLIMIT_NOFILE, (soft, hard))
             except Exception:
-                default_logger.warning('failed to set ulimit, giving up')
+                print('failed to set ulimit, giving up')
                 soft, hard = res.getrlimit(res.RLIMIT_NOFILE)
 
-    default_logger.debug(f'ulimit -n soft,hard: {soft} {hard}')
     return soft, hard
 
 
 _set_nofile()
 
-# Flow
-from jina.flow import Flow
-from jina.flow.asyncio import AsyncFlow
+# ONLY FIRST CLASS CITIZENS ARE ALLOWED HERE, namely Document, Executor Flow
 
 # Client
 from jina.clients import Client
-from jina.clients.asyncio import AsyncClient
+
+# Document
+from jina.types.document import Document
+from jina.types.arrays.document import DocumentArray
 
 # Executor
-from jina.executors import GenericExecutor as Executor
-from jina.executors.classifiers import BaseClassifier as Classifier
-from jina.executors.crafters import BaseCrafter as Crafter
-from jina.executors.encoders import BaseEncoder as Encoder
-from jina.executors.evaluators import BaseEvaluator as Evaluator
-from jina.executors.indexers import BaseIndexer as Indexer
-from jina.executors.rankers import BaseRanker as Ranker
-from jina.executors.segmenters import BaseSegmenter as Segmenter
+from jina.executors import BaseExecutor as Executor
 from jina.executors.decorators import requests
 
+# Flow
+from jina.flow.base import Flow
+from jina.flow.asyncio import AsyncFlow
+
 __all__ = [_s for _s in dir() if not _s.startswith('_')]
-__all__.extend([_s for _s in _names_with_underscore])
+__all__.extend(_names_with_underscore)
