@@ -270,6 +270,10 @@ class JAML:
                 # "root" context is now the global namespace
                 # "this" context is now the current node namespace
                 v = v.format(root=expand_map, this=p, ENV=env_map)
+            except AttributeError as ex:
+                raise AttributeError(
+                    'variable replacement is failed, please check your YAML file.'
+                ) from ex
             except KeyError:
                 pass
 
@@ -430,7 +434,8 @@ class JAMLCompatible(metaclass=JAMLCompatibleType):
         f = filename or getattr(self, 'config_abspath', None)
         if not f:
             f = tempfile.NamedTemporaryFile(
-                'w', delete=False, dir=os.environ.get('JINA_EXECUTOR_WORKDIR', None)
+                'w',
+                delete=False,
             ).name
             warnings.warn(
                 f'no "filename" is given, {self!r}\'s config will be saved to: {f}'
@@ -445,7 +450,9 @@ class JAMLCompatible(metaclass=JAMLCompatibleType):
         *,
         allow_py_modules: bool = True,
         substitute: bool = True,
-        context: Dict[str, Any] = None,
+        context: Optional[Dict[str, Any]] = None,
+        override_with: Optional[Dict] = None,
+        override_metas: Optional[Dict] = None,
         **kwargs,
     ) -> 'JAMLCompatible':
         """A high-level interface for loading configuration with features
@@ -492,6 +499,8 @@ class JAMLCompatible(metaclass=JAMLCompatibleType):
         :param allow_py_modules: allow importing plugins specified by ``py_modules`` in YAML at any levels
         :param substitute: substitute environment, internal reference and context variables.
         :param context: context replacement variables in a dict, the value of the dict is the replacement.
+        :param override_with: dictionary of parameters to overwrite from the default config
+        :param override_metas: dictionary of parameters to overwrite from the default config
         :param kwargs: kwargs for parse_config_source
         :return: :class:`JAMLCompatible` object
         """
@@ -500,8 +509,22 @@ class JAMLCompatible(metaclass=JAMLCompatibleType):
             # first load yml with no tag
             no_tag_yml = JAML.load_no_tags(fp)
             if no_tag_yml:
-                # extra arguments are parsed to inject_config
-                no_tag_yml = cls.inject_config(no_tag_yml, **kwargs)
+                no_tag_yml.update(**kwargs)
+                if override_with is not None:
+                    with_params = no_tag_yml.get('with', None)
+                    if with_params:
+                        with_params.update(**override_with)
+                        no_tag_yml.update(with_params)
+                    else:
+                        no_tag_yml['with'] = override_with
+                if override_metas is not None:
+                    metas_params = no_tag_yml.get('metas', None)
+                    if metas_params:
+                        metas_params.update(**override_metas)
+                        no_tag_yml.update(metas_params)
+                    else:
+                        no_tag_yml['metas'] = override_metas
+
             else:
                 raise BadConfigSource(
                     f'can not construct {cls} from an empty {source}. nothing to read from there'
@@ -515,31 +538,16 @@ class JAMLCompatible(metaclass=JAMLCompatibleType):
                     no_tag_yml,
                     extra_search_paths=(os.path.dirname(s_path),) if s_path else None,
                 )
-            from ..flow import BaseFlow
+            from ..flow.base import Flow
 
-            if issubclass(cls, BaseFlow):
+            if issubclass(cls, Flow):
                 tag_yml = JAML.unescape(
                     JAML.dump(no_tag_yml),
                     include_unknown_tags=False,
-                    jtype_whitelist=('Flow', 'AsyncFlow'),
+                    jtype_whitelist=('Flow',),
                 )
             else:
                 # revert yaml's tag and load again, this time with substitution
                 tag_yml = JAML.unescape(JAML.dump(no_tag_yml))
-
             # load into object, no more substitute
             return JAML.load(tag_yml, substitute=False)
-
-    @classmethod
-    def inject_config(cls, raw_config: Dict, *args, **kwargs) -> Dict:
-        """Inject/modify the config before loading it into an object.
-
-         .. note::
-            This function is most likely to be overridden by its subclass.
-
-        :param raw_config: raw config to work on
-        :param args: not used
-        :param kwargs: not used
-        :return: the config
-        """
-        return raw_config
